package httpserver

import (
	"context"
	"crypto/sha256"
	"encoding/base64"
	"fmt"
	"io"
	"log"
	"math/rand"
	"net/http"
	"path/filepath"
	"text/template"

	"git.netflux.io/rob/elon-eats-my-tweets/config"
	"github.com/go-chi/chi"
	"github.com/go-chi/chi/middleware"
	"github.com/gorilla/sessions"
	"golang.org/x/oauth2"
)

const (
	sessionName            = "elon_session"
	sessionKeyState        = "state"
	sessionKeyPkceVerifier = "pkce_verifier"
	stateLen               = 64
	pkceVerifierLen        = 64
)

type handler struct {
	templates    *template.Template
	oauth2Config *oauth2.Config
	sessionStore *sessions.CookieStore
}

func newHandler(cfg config.Config, templates *template.Template) *handler {
	return &handler{
		templates:    templates,
		sessionStore: sessions.NewCookieStore([]byte(cfg.SessionKey)),
		oauth2Config: &oauth2.Config{
			ClientID:     cfg.Twitter.ClientID,
			ClientSecret: cfg.Twitter.ClientSecret,
			RedirectURL:  cfg.Twitter.CallbackURL,
			Scopes:       []string{"tweet.read", "tweet.write", "users.read", "offline.access"},
			Endpoint: oauth2.Endpoint{
				AuthURL:   "https://twitter.com/i/oauth2/authorize",
				TokenURL:  "https://api.twitter.com/2/oauth2/token",
				AuthStyle: oauth2.AuthStyleInHeader,
			},
		},
	}
}

func (h *handler) getIndex(w http.ResponseWriter, r *http.Request) {
	if err := h.templates.ExecuteTemplate(w, "index", nil); err != nil {
		log.Printf("error rendering template: %v", err)
		http.Error(w, "error rendering template", http.StatusInternalServerError)
		return
	}
}

func (h *handler) getLogin(w http.ResponseWriter, r *http.Request) {
	state := randSeq(stateLen)
	pkceVerifier := randSeq(pkceVerifierLen)

	session, _ := h.sessionStore.Get(r, sessionName)
	session.Values[sessionKeyState] = state
	session.Values[sessionKeyPkceVerifier] = pkceVerifier
	if err := session.Save(r, w); err != nil {
		log.Printf("error saving session: %v", err)
		http.Error(w, "error saving session", http.StatusBadRequest)
		return
	}

	url := h.oauth2Config.AuthCodeURL(
		state,
		oauth2.SetAuthURLParam("code_challenge", encodeSHA256(pkceVerifier)),
		oauth2.SetAuthURLParam("code_challenge_method", "S256"),
	)

	http.Redirect(w, r, url, http.StatusTemporaryRedirect)
}

func (h *handler) getCallback(w http.ResponseWriter, r *http.Request) {
	session, err := h.sessionStore.Get(r, sessionName)
	if err != nil {
		log.Printf("error reading session: %v", err)
		http.Error(w, "error reading session", http.StatusBadRequest)
		return
	}

	state, ok := session.Values[sessionKeyState]
	if !ok {
		log.Println("empty state", err)
		http.Error(w, "error reading session", http.StatusBadRequest)
		return
	}

	if state != r.URL.Query().Get("state") {
		http.Error(w, "error validating request", http.StatusBadRequest)
		return
	}

	code := r.URL.Query().Get("code")
	if code == "" {
		http.Error(w, "empty code", http.StatusBadRequest)
		return
	}

	pkceVerifier, ok := session.Values[sessionKeyPkceVerifier]
	if !ok {
		log.Println("empty code challenge", err)
		http.Error(w, "error reading session", http.StatusBadRequest)
		return
	}

	token, err := h.oauth2Config.Exchange(context.Background(), code, oauth2.SetAuthURLParam("code_verifier", pkceVerifier.(string)))
	if err != nil {
		log.Printf("error exchanging code: %v", err)
		http.Error(w, "error exchanging code", http.StatusForbidden)
		return
	}

	client := h.oauth2Config.Client(context.Background(), token)
	resp, err := client.Get("https://api.twitter.com/2/users/me")
	if err != nil {
		log.Printf("error fetching users/me: %v", err)
		http.Error(w, "error fetching user", http.StatusInternalServerError)
		return
	}
	defer resp.Body.Close()
	// TODO: do something sensible
	body, _ := io.ReadAll(resp.Body)

	w.Header().Set("content-type", "application/json")
	w.WriteHeader(http.StatusOK)
	if _, err = w.Write([]byte(body)); err != nil {
		log.Printf("error writing response: %v", err)
	}
}

func Start(cfg config.Config) error {
	r := chi.NewRouter()
	r.Use(middleware.RequestID)
	r.Use(middleware.RealIP)
	r.Use(middleware.Logger)
	r.Use(middleware.Recoverer)

	templates, err := template.ParseGlob(filepath.Join(cfg.PublicPath, "views", "*.html"))
	if err != nil {
		return fmt.Errorf("error loading templates: %v", err)
	}

	h := newHandler(cfg, templates)
	r.Get("/", h.getIndex)
	r.Get("/login", h.getLogin)
	r.Get("/callback", h.getCallback)

	return http.ListenAndServe(":8000", r)
}

// https://stackoverflow.com/questions/22892120/how-to-generate-a-random-string-of-a-fixed-length-in-go
var letters = []rune("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ")

func randSeq(n int) string {
	b := make([]rune, n)
	for i := range b {
		b[i] = letters[rand.Intn(len(letters))]
	}
	return string(b)
}

func encodeSHA256(s string) string {
	enc := sha256.Sum256([]byte(s))
	return base64.RawURLEncoding.EncodeToString(enc[:])
}
